﻿// Atac Template Engine
// Copyright (C) 2011 Markus Jahn <markus.jahn@gmx.at>
//
// This file is part of Atac Template Engine.
//
// Atac Template Engine is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Atac Template Engine is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Atac Template Engine.  If not, see <http://www.gnu.org/licenses/>.

using NUnit.Framework;

namespace Atac.TemplateEngine.Tests {
  internal class BucketNodeTests : TestCase<INode> {
    [Test]
    public void TextTemplate() {
      Init(@"
UnitDecl =
namespace name { }
.
");
      Node ast = new Node(NodeKind.UnitDecl);
      GenerateAndVerify(ast, "namespace name { }", new BucketTreeHelper());
    }


    [Test]
    public void EmptyTemplate() {
      Init(@"
ProgramDecl =

.
");
      Node ast = new Node(NodeKind.ProgramDecl);
      GenerateAndVerify(ast, "", new BucketTreeHelper());
    }


    [Test]
    public void LineFeed() {
      Init(@"
ProgramDecl =


.
");
      Node ast = new Node(NodeKind.ProgramDecl);
      GenerateAndVerify(ast, "\r\n", new BucketTreeHelper());
    }


    [Test]
    public void SepInText() {
      Init(@"
ProgramDecl =
Text sep Text TextsepTtext sep Text
  sep
.
");
      Node ast = new Node(NodeKind.ProgramDecl);
      GenerateAndVerify(ast, @"
Text sep Text TextsepTtext sep Text
  sep".TrimStart(), new BucketTreeHelper());
    }


    [Test]
    public void SpecialCharacters() {
      Init(@"
ProgramDecl =
\<Program\>
  Text /\/ .= 4 \""
\</Program\>
.
");
      Node ast = new Node(NodeKind.ProgramDecl);
      GenerateAndVerify(ast, @"
<Program>
  Text // .= 4 ""
</Program>".TrimStart(), new BucketTreeHelper());
    }


    [Test]
    public void NodeElement() {
      Init(@"
VarDecl(Type, VarId(Expr)) =
<Type> <VarId>< "" = "" Expr>;
.
VarId=
name
.
Integer=
int
.
IntVal =
3
.
");
      Node ast = new Node(NodeKind.VarDecl) {
          Left = new Node(NodeKind.Integer),
          Right = new Node(NodeKind.VarId) {
              Left = new Node(NodeKind.IntVal)
          }
      };
      GenerateAndVerify(ast, @"
int name = 3;".TrimStart(), new BucketTreeHelper());
    }


    [Test]
    public void NodeElements() {
      Init(@"
VarDecl(Type, VarId(Expr)) =
<Type> <""prec "" VarId sep "", +<""+""VarId sep""-""""+"">+"" "" succ"">;
.
VarId=
name
.
Integer=
int
.
");
      Node ast = new Node(NodeKind.VarDecl) {
          Left = new Node(NodeKind.Integer),
          Right = new Node(NodeKind.VarId) {
              Next = new Node(NodeKind.VarId) {
                  Next = new Node(NodeKind.VarId)
              }
          }
      };
      GenerateAndVerify(ast, @"
int prec name, ++name-name-name++name, ++name-name-name++name succ;
".Trim(), new BucketTreeHelper());
    }


    [Test]
    public void Indentation() {
      Init(@"
ClassType(Decls, BaseTypes) =
class T1 < "":"" BaseTypes sep"", "" > {
  <Decls>
}
.
VarDecl(Type, VarId(Expr))[1,2] =
  <Type>
  <VarId sep "", "">;
.
VarId=
name
.
Integer=
int
.
");
      Node ast = new Node(NodeKind.ClassType) {
        Left = new Node(NodeKind.VarDecl) {
          Left = new Node(NodeKind.Integer),
          Right = new Node(NodeKind.VarId),
          Next = new Node(NodeKind.VarDecl) {
              Left = new Node(NodeKind.Integer),
              Right = new Node(NodeKind.VarId)
          }
        }
      };
      GenerateAndVerify(ast, @"
class T1  {
  
    int
    name;
  
    int
    name;

}
".Trim(), new BucketTreeHelper());
    }
  }
}