﻿// Atac Template Engine
// Copyright (C) 2011 Markus Jahn <markus.jahn@gmx.at>
//
// This file is part of Atac Template Engine.
//
// Atac Template Engine is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Atac Template Engine is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Atac Template Engine.  If not, see <http://www.gnu.org/licenses/>.

using NUnit.Framework;

namespace Atac.TemplateEngine.Tests {
  internal class TemplateParserTests : TestCase<INode> {
    [Test]
    public void EmptyTemplate() {
      expectedTemplates.Add("UnitDecl", new Template("UnitDecl"));
      Init(@"
UnitDecl = 
.
");
      ParseAndVerifyTemplates();
    }

    [Test]
    public void WhitspaceAfterLastTemplate() {
      expectedTemplates.Add("UnitDecl", new Template("UnitDecl"));
      Init(@"
UnitDecl = 
.


");
      ParseAndVerifyTemplates();
    }

    [Test]
    public void EmptyNodeDecl() {
      expectedTemplates.Add("UnitDecl", new Template(
          "UnitDecl", new[] { PathDecl.Empty }));
      Init(@"
UnitDecl() = 
.
");
      ParseAndVerifyTemplates();
    }

    [Test]
    public void MultipleEmptyNodeDecl() {
      expectedTemplates.Add("UnitDecl", new Template(
          "UnitDecl", new[] { PathDecl.Empty, PathDecl.Empty, PathDecl.Empty }
          ));
      Init(@"
UnitDecl(,,) = 
.
");
      ParseAndVerifyTemplates();
    }

    [Test]
    public void SingleNodeDecl() {
      expectedTemplates.Add("UnitDecl", new Template(
          "UnitDecl", new[] { new PathDecl("Decl") }
          ));
      Init(@"
UnitDecl( Decl ) =
.
");
      ParseAndVerifyTemplates();
    }

    [Test]
    public void MultipleNodeDecls() {
      expectedTemplates.Add("UnitDecl", new Template(
          "UnitDecl", new[] { new PathDecl("Decl"), new PathDecl("Using") }
          ));
      Init(@"
UnitDecl(Decl, Using) =
.
");
      ParseAndVerifyTemplates();
    }

    [Test]
    public void SkipNodeDecl() {
      expectedTemplates.Add("UnitDecl", new Template(
          "UnitDecl", new[] { PathDecl.Empty, new PathDecl("Using") }));
      Init(@"
UnitDecl(, Using) =
.
");
      ParseAndVerifyTemplates();
    }

    [Test]
    public void NestedNodeDecls() {
      expectedTemplates.Add("UnitDecl", new Template(
          "UnitDecl", new[] {
              new PathDecl("Decl"),
              new PathDecl("Using", new[] { new PathDecl("Unit"), new PathDecl("StrVal") })
          }));
      Init(@"
UnitDecl(Decl, Using(Unit,StrVal)) =
.
");
      ParseAndVerifyTemplates();
    }

    [Test]
    public void DuplicateNodeDecl() {
      expectedTemplates.Add("UnitDecl", new Template(
          "UnitDecl", new[] {
              new PathDecl("Decl"),
              new PathDecl("Using", new[] { new PathDecl("Unit"), new PathDecl("Decl") })
          }));
      expectedErrors.Add(new Log("Path 'Decl' already declared", 2, 27));
      Init(@"
UnitDecl(Decl, Using(Unit,Decl)) =
.
");
      ParseAndVerifyTemplates();
    }

    [Test]
    public void EqualNodeDeclAsTemplateName() {
      expectedTemplates.Add("UnitDecl", new Template(
          "UnitDecl", new[] {
              new PathDecl("Decl"), new PathDecl("UnitDecl")
          }));
      expectedErrors.Add(new Log("Path 'UnitDecl' already declared", 2, 16));
      Init(@"
UnitDecl(Decl, UnitDecl) =
.
");
      ParseAndVerifyTemplates();
    }

    [Test]
    public void InvalidLineFeedDecl1() {
      expectedTemplates.Add("UnitDecl", new Template("UnitDecl"));
      expectedErrors.Add(new Log("invalid LineFeedDecl", 2, 11));
      Init(@"
UnitDecl [] =
.
");
      ParseAndVerifyTemplates();
    }

    [Test]
    public void InvalidLineFeedDecl2() {
      expectedTemplates.Add("UnitDecl", new Template("UnitDecl"));
      expectedErrors.Add(new Log("number expected", 2, 12));
      Init(@"
UnitDecl [,] =
.
");
      ParseAndVerifyTemplates();
    }

    [Test]
    public void InvalidLineFeedDecl3() {
      expectedTemplates.Add("UnitDecl",
          new Template("UnitDecl") { SucceedingLineFeeds = 1 });
      expectedErrors.Add(new Log("\"]\" expected", 2, 12));
      Init(@"
UnitDecl[,1,1] =
.
");
      ParseAndVerifyTemplates();
    }

    [Test]
    public void PrecedingLineFeedDecl() {
      expectedTemplates.Add("UnitDecl",
          new Template("UnitDecl") { PrecedingLineFeeds = 1 });
      Init(@"
UnitDecl[1] =
.
");
      ParseAndVerifyTemplates();
    }

    [Test]
    public void SucceedingLineFeedDecl() {
      expectedTemplates.Add("UnitDecl",
          new Template("UnitDecl") { SucceedingLineFeeds = 3 });
      Init(@"
UnitDecl[,3] =
.
");
      ParseAndVerifyTemplates();
    }

    [Test]
    public void LineFeedDecl() {
      expectedTemplates.Add("UnitDecl",
          new Template("UnitDecl") {
              PrecedingLineFeeds = 2,
              SucceedingLineFeeds = 3
          });
      Init(@"
UnitDecl[2,3] =
.
");
      ParseAndVerifyTemplates();
    }

    [Test]
    public void NewLine() {
      Template template = new Template("UnitDecl");
      template.Expression = new Expression(
        new[] { new Line(), new Line() });
      expectedTemplates.Add("UnitDecl",
          template);
      Init(@"
UnitDecl =


.
");
      ParseAndVerifyTemplates();
    }

    [Test]
    public void SimplePlaceholder() {
      Template template = new Template("VarDecl");
      template.Expression = new Expression(
        new[] { new Line(new[] { new Placeholder("Type") }) });
      expectedTemplates.Add("VarDecl", template);
      Init(@"
VarDecl =
<Type>
.
");
      ParseAndVerifyTemplates();
    }

    [Test]
    public void TemplateLink() {
      Template template = new Template("VarId");
      template.Expression = new Expression(
        new[] { new Line(new[] { new Placeholder(new PathItem[0]) { Link = "Ident" } } ) });
      expectedTemplates.Add("VarId", template);
      Init(@"
VarId =
<VarId:Ident>
.
");
      ParseAndVerifyTemplates();
    }
  }
}