﻿// Atac Template Engine
// Copyright (C) 2011 Markus Jahn <markus.jahn@gmx.at>
//
// This file is part of Atac Template Engine.
//
// Atac Template Engine is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Atac Template Engine is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Atac Template Engine.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using NUnit.Framework;

namespace Atac.TemplateEngine.Tests {
  [TestFixture]
  internal abstract class TestCase<TNode> where TNode : class {
    protected TemplateEngine<TNode> templateEngine;
    protected StringWriter writer;
    //protected MemoryStream templateStream, templateHandlerStream;
    protected TestLogger logger;

    protected string templateString, templateHandlerString, output;
    protected TNode root;

    protected Dictionary<string, Template> expectedTemplates;
    protected List<Log> expectedErrors;

    protected void Init(string tempaltes) {
      Init(tempaltes, string.Empty);
    }

    protected void Init(string templates, string templateHandlers) {
      templateString = templates;
      templateHandlerString = templateHandlers;
      Stream templateStream = new MemoryStream(Encoding.UTF8.GetBytes(templates));
      Stream templateHandlerStream = new MemoryStream(Encoding.UTF8.GetBytes(templateHandlers));
      templateEngine = new TemplateEngine<TNode>(templateStream, templateHandlerStream);
      logger = new TestLogger();
      templateEngine.Logger = logger;
    }

    protected void Init(string templates, Stream templateHandlers) {
      Init(templates, new StreamReader(templateHandlers).ReadToEnd());
    }

    protected void Init(string templates, object templateHandlerObject) {
      Init(templates, string.Empty);
      templateEngine.UpdateTemplateHandlers(templateHandlerObject);
    }

    private void PrintResult() {
      if (output != null) {
        Console.WriteLine("\n--- Output ---------------");
        Console.WriteLine(output);
      }
      if (!string.IsNullOrEmpty(templateString)) {
        Console.WriteLine("\n--- Template(s) ----------");
        Console.WriteLine(templateString.Trim());
      }
      if (!string.IsNullOrEmpty(templateHandlerString)) {
        Console.WriteLine("\n--- TemplateHandler(s) ---");
        Console.WriteLine(templateHandlerString.Trim());
      }
      //if (root != null) {
      //  Console.WriteLine("\n--- Ast ------------------");
      //  PrintNode(root);
      //  Console.WriteLine();
      //}
    }

    //private static void PrintNode(TNode node) {
    //  if (node == null) {
    //    return;
    //  }
    //  Console.Write(node);
    //  if (node.Children.Count > 0) {
    //    Console.Write("(");
    //    Console.Write(node.Children[0]);
    //    for(int i = 1; i < node.Children.Count; i++) {
    //      Console.Write(",");
    //      Console.Write(node.Children[i]);
    //    }
    //    Console.Write(")");
    //  }
    //}

    protected void VerifyErrors() {
      List<Log> actualErrors = logger.Errors;
      Assert.AreEqual(expectedErrors.Count, actualErrors.Count, "Error count");
      for (int i = 0; i < expectedErrors.Count; i++) {
        VerifyLog(expectedErrors[i], actualErrors[i], "Error");
      }
    }

    protected void VerifyLog(Log expected, Log actual, string logKind) {
      Assert.AreEqual(expected.Message, actual.Message, logKind + " message");
      Assert.AreEqual(expected.Line, actual.Line, logKind + " line");
      Assert.AreEqual(expected.Col, actual.Col, logKind + " column");
    }

    protected void ParseAndVerifyTemplates() {
      templateEngine.UpdateTemplates();
      templateEngine.UpdateTemplateHandlers();
      VerifyErrors();
      if (logger.Errors.Count > 0) {
        return;
      }
      Dictionary<string, Template> actualTemplates = templateEngine.Templates;
      Assert.AreEqual(expectedTemplates.Count, actualTemplates.Count);
      foreach (string key in expectedTemplates.Keys) {
        Assert.Contains(key, actualTemplates.Keys);
        Template expectedTemplate = expectedTemplates[key];
        Template actualTemplate = actualTemplates[key];
        VerifyTemplates(expectedTemplate, actualTemplate);
      }
    }

    protected void VerifyTemplates(Template expected, Template actual) {
      Assert.AreEqual(expected.Name, actual.Name);
      Assert.AreEqual(expected.PrecedingLineFeeds, actual.PrecedingLineFeeds);
      Assert.AreEqual(expected.SucceedingLineFeeds, actual.SucceedingLineFeeds);
      VerifyPathDecls(expected.PathDecls, actual.PathDecls);
      VerifyExpression(expected.Expression, actual.Expression);
    }

    private void VerifyPathDecls(IList<PathDecl> expected, IList<PathDecl> actual) {
      if (expected == null) {
        Assert.Null(actual);
      } else {
        Assert.NotNull(actual);
        Assert.AreEqual(expected.Count, actual.Count);
        for (int i = 0; i < expected.Count; i++) {
          VerifyPathDecls(expected[i], actual[i]);
        }
      }
    }

    protected void VerifyPathDecls(PathDecl expected, PathDecl actual) {
      Assert.AreEqual(expected.Name, actual.Name);
      if (expected.Children != null) {
        Assert.NotNull(actual.Children);
        Assert.AreEqual(expected.Children.Count, actual.Children.Count);
        for (int i = 0; i < expected.Children.Count; i++) {
          VerifyPathDecls(expected.Children[i], actual.Children[i]);
        }
      } else {
        Assert.Null(actual.Children);
      }
    }

    protected void VerifyExpression(Expression expected, Expression actual) {
      if (expected == null) {
        Assert.Null(actual);
      } else {
        Assert.NotNull(actual);
        VerifyLines(expected.Lines, actual.Lines);
      }
    }

    protected void VerifyLines(List<Line> expected, List<Line> actual) {
      if (expected == null) {
        Assert.Null(actual);
      } else {
        Assert.NotNull(actual);
        Assert.AreEqual(expected.Count, actual.Count);
        for(int i = 0; i < expected.Count; i++) {
          VerifyLine(expected[i], actual[i]);
        }
      }
    }

    protected void VerifyLine(Line expected, Line actual) {
      if (expected == null) {
        Assert.Null(actual);
      } else {
        Assert.NotNull(actual);
        Assert.AreEqual(expected.Indentation, actual.Indentation);
        VerifyLineElements(expected.Elements, actual.Elements);
      }
    }

    protected void VerifyLineElements(List<LineElement> expected, List<LineElement> actual) {
      if (expected == null) {
        Assert.Null(actual);
      } else {
        Assert.NotNull(actual);
        Assert.AreEqual(expected.Count, actual.Count);
        for(int i = 0; i < expected.Count; i++) {
          VerifyLineElement(expected[i], actual[i]);
        }
      }
    }

    protected void VerifyLineElement(LineElement expected, LineElement actual) {
      Assert.AreEqual(expected.GetType(), actual.GetType());
      if (expected is Text) {
        VerifyTextElement(expected, actual);
      } else if (expected is Placeholder) {
        VerifyPlaceholder((Placeholder)expected, (Placeholder)actual);
      }
    }

    private static void VerifyTextElement(LineElement expected, LineElement actual) {
      Assert.AreEqual(((Text) expected).Value, ((Text) actual).Value);
    }

    protected void VerifyPlaceholder(Placeholder expected, Placeholder actual) {
      if (expected.Path == null) {
        Assert.Null(actual.Path);
      } else {
        Assert.NotNull(actual.Path);
        Assert.AreEqual(expected.Path.Count, actual.Path.Count);
        for(int i = 0; i < expected.Path.Count; i++) {
          VerifyPathItem(expected.Path[i], actual.Path[i]);
        }
      }
    }

    protected void VerifyPathItem(PathItem expected, PathItem actual) {
      Assert.AreEqual(expected.Name, actual.Name);
      Assert.AreEqual(expected.Index, actual.Index);
    }

    //class PathItemComparer : IComparer {
    //  public int Compare(PathItem x, PathItem y) {
    //    throw new NotImplementedException();
    //  }

    //  public int Compare(object x, object y) {
    //    throw new NotImplementedException();
    //  }
    //}

    protected void GenerateAndVerify(TNode node, string expected, TreeHelper<TNode> treeHelper) {
      templateEngine.TreeHelper = treeHelper;
      GenerateAndVerify(node, expected);
    }

    protected void GenerateAndVerify(TNode node, string expected) {
      root = node;
      writer = new StringWriter();
      templateEngine.Start(writer, root);
      output = writer.ToString();
      Assert.AreEqual(expected, output);
    }

    [SetUp]
    public void SetUp() {
      expectedTemplates = new Dictionary<string, Template>();
      expectedErrors = new List<Log>();
      root = null;
      output = null;
    }

    [TearDown]
    public void TearDown() {
      PrintResult();
    }
  }
}