﻿// Atac Template Engine
// Copyright (C) 2011 Markus Jahn <markus.jahn@gmx.at>
//
// This file is part of Atac Template Engine.
//
// Atac Template Engine is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Atac Template Engine is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Atac Template Engine.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.IO;
using System.Reflection;

namespace Atac.TemplateEngine {
  public delegate void NodeEventHandler<TNode>(object sender, NodeEventArgs<TNode> args)
      where TNode : class;

  public class CodeGenerator<TNode> where TNode : class {
    internal const MemberTypes MEMBER_TYPES = MemberTypes.Field | MemberTypes.Property;
    internal const BindingFlags BINDING_FLAGS = BindingFlags.Static | BindingFlags.Instance
        | BindingFlags.Public; // | BindingFlags.NonPublic;

    private int dumpedLineFeeds;
    private int pendingLineFeeds;
    private readonly Stack<string> indentationStack;

    public event NodeEventHandler<TNode> BeforePutNode;
    public event NodeEventHandler<TNode> AfterPutNode;

    public TextWriter Writer { get; private set; }
    public TreeHelper<TNode> TreeHelper { get; private set; }
    public Dictionary<string, Template> Templates { get; private set; }
    public Dictionary<string, TemplateHandler<TNode>> TemplateHandlers { get; private set; }
    public ILogger Logger { get; private set; }

    private readonly Stack<NodeItem> nodeStack;

    internal int DumpedLineFeeds {
      get { return dumpedLineFeeds; }
    }

    internal int PendingLineFeeds {
      get { return pendingLineFeeds; }
      set {
        if (value < 0) {
          throw new ArgumentException("invalid value");
        }
        pendingLineFeeds = value;
      }
    }

    public TNode CurrentNode {
      get { return nodeStack.Peek().Node; }
    }

    public CodeGenerator(
        TextWriter writer,
        TreeHelper<TNode> treeHelper,
        Dictionary<string, Template> templates,
        Dictionary<string, TemplateHandler<TNode>> templateHandlers,
        ILogger logger) {
      if (writer == null) {
        throw new ArgumentNullException("writer");
      }
      if (treeHelper == null) {
        throw new ArgumentNullException("treeHelper");
      }
      if (templates == null) {
        throw new ArgumentNullException("templates");
      }
      if (templateHandlers == null) {
        throw new ArgumentNullException("templateHandlers");
      }
      if (logger == null) {
        throw new ArgumentNullException("logger");
      }
      Writer = writer;
      TreeHelper = treeHelper;
      Templates = templates;
      TemplateHandlers = templateHandlers;
      Logger = logger;
      nodeStack = new Stack<NodeItem>();
      indentationStack = new Stack<string>();
      indentationStack.Push(string.Empty);

      treeHelper.Init(this);
    }

    #region Public Methods

    public bool PutText(object value) {
      if (value == null) {
        throw new ArgumentNullException("value");
      }
      return InternalPutText(value);
    }

    public bool PutNode(TNode node) {
      if (node == null) {
        throw new ArgumentNullException("node");
      }
      return InternalPutNode(node, (string) null);
    }

    public bool PutNode(TNode node, string template) {
      if (node == null) {
        throw new ArgumentNullException("node");
      }
      return InternalPutNode(node, template);
    }

    public bool PutTemplate(TNode node) {
      if (node == null) {
        throw new ArgumentNullException("node");
      }
      return InternalPutTemplate(node, string.Empty);
    }

    public bool PutTemplate(TNode node, string template) {
      if (node == null) {
        throw new ArgumentNullException("node");
      }
      return InternalPutTemplate(node, template);
    }

    public bool PutNewLine() {
      return InternalPutNewLine(false);
    }

    public bool PutNewLine(bool ignoreIndentation) {
      return InternalPutNewLine(ignoreIndentation);
    }

    public void PutIndentation(object indent) {
      if (indent == null) {
        throw new ArgumentNullException("indent");
      }
      InternalPutIndentation(indent);
    }

    public void ReduceIndentation() {
      InternalReduceIndentation();
    }

    #endregion

    #region Logger Helper

    public void LogError(string msg, params object[] args) {
      if (Logger != null) {
        Logger.LogError(string.Format(msg, args));
      }
    }

    public void LogError(int line, int col, string msg, params object[] args) {
      if (Logger != null) {
        Logger.LogError(string.Format(msg, args), line, col);
      }
    }

    public void LogWarning(string msg, params object[] args) {
      if (Logger != null) {
        Logger.LogWarning(string.Format(msg, args));
      }
    }

    public void LogInfo(int line, int col, string msg, params object[] args) {
      if (Logger != null) {
        Logger.LogInfo(string.Format(msg, args), line, col);
      }
    }

    public void LogInfo(string msg, params object[] args) {
      if (Logger != null) {
        Logger.LogInfo(string.Format(msg, args));
      }
    }

    public void LogWarning(int line, int col, string msg, params object[] args) {
      if (Logger != null) {
        Logger.LogWarning(string.Format(msg, args), line, col);
      }
    }

    private void InternalLogError(string nodeKind, string templateKind, string templateName, string message) {
      if (templateKind == null) {
        LogError("Error on putting node '{0}': {1}", nodeKind, message);
      } else if (nodeKind == templateName || string.IsNullOrEmpty(templateName)) {
        LogError("Error on putting {0} for node '{1}': {2}", templateKind, nodeKind, message);
      } else {
        LogError("Error on putting {0} '{1}' for node '{2}': {3}", templateKind, templateName, nodeKind, message);
      }
    }

    #endregion

    #region Internal Methods

    internal bool InternalPutText(object value) {
      string text = value.ToString();
      if (text == string.Empty) {
        return false;
      }
      while (pendingLineFeeds > 0) {
        Writer.WriteLine();
        Writer.Write(indentationStack.Peek());
        dumpedLineFeeds++;
        pendingLineFeeds--;
      }
      //PutPendingElements();
      //if (precedingElements != null) {
      //  foreach (ITemplateElement precedingElement in precedingElements) {
      //    precedingElement.Put(this);
      //  }
      //}
      Writer.Write(text);
      dumpedLineFeeds = 0;
      return true;
    }

    internal bool InternalPutNewLine(bool ignoreIndentation) {
      Writer.WriteLine();
      dumpedLineFeeds++;
      if (pendingLineFeeds > 0) {
        pendingLineFeeds--;
      }
      if (!ignoreIndentation) {
        Writer.Write(indentationStack.Peek());
      }
      return true;
    }

    internal bool InternalPutNode(TNode node, string templateName) {
      bool isTemplateSpecified;
      string nodeKind = TreeHelper.GetNodeKind(node);
      if (string.IsNullOrEmpty(templateName)) {
        templateName = nodeKind;
        isTemplateSpecified = false;
      } else {
        isTemplateSpecified = true;
      }
      if (TemplateHandlers.ContainsKey(templateName)) {
        return InternalPutNode(node, TemplateHandlers[templateName]);
      }
      if (Templates.ContainsKey(templateName)) {
        return InternalPutNode(node, Templates[templateName]);
      }
      if (!isTemplateSpecified) {
        return InternalPutNode(node);
      }
      InternalLogError(nodeKind, "template", templateName, "Template not found");
      return false;
    }

    internal bool InternalPutTemplate(TNode node, string templateName) {
      string nodeKind = TreeHelper.GetNodeKind(node);
      templateName = string.IsNullOrEmpty(templateName) ? nodeKind : templateName;
      if (Templates.ContainsKey(templateName)) {
        return InternalPutNode(node, Templates[templateName]);
      }
      InternalLogError(nodeKind, "template", templateName, "Template not found");
      return false;
    }

    private bool InternalPutNode(TNode node, TemplateHandler<TNode> templateHandler) {
      NodeItem nodeItem = new NodeItem(node, templateHandler);
      if (nodeStack.Contains(nodeItem)) {
        InternalLogError(
            TreeHelper.GetNodeKind(node), "template handler", templateHandler.Method.Name,
            "Recursively calling the same template handler for a node is not allowed");
        return false;
      }
      OnBeforePutNode(nodeItem);
      try {
        templateHandler(node, this);
        return true;
      } catch (Exception e) {
        InternalLogError(
            TreeHelper.GetNodeKind(node), "template handler", templateHandler.Method.Name,
            e.Message);
        return false;
      } finally {
        OnAfterPutNode(nodeItem);
      }
    }

    private bool InternalPutNode(TNode node, Template template) {
      NodeItem nodeItem = new NodeItem(node, template);
      if (nodeStack.Contains(nodeItem)) {
        InternalLogError(TreeHelper.GetNodeKind(node), "template", template.Name,
            "Recursively applying the same template for a node is not allowed");
        return false;
      }
      OnBeforePutNode(nodeItem);
      template.Put(node, this);
      OnAfterPutNode(nodeItem);
      return true;
    }

    private bool InternalPutNode(TNode node) {
      NodeItem nodeItem = new NodeItem(node, null);
      if (nodeStack.Contains(nodeItem)) {
        InternalLogError(TreeHelper.GetNodeKind(node), null, null,
            "Recursively putting the same node is not allowed");
        return false;
      }
      OnBeforePutNode(nodeItem);
      InternalPutText(node);
      OnAfterPutNode(nodeItem);
      return true;
    }

    internal bool InternalPutIndentation(object indent) {
      bool result = InternalPutText(indent);
      indentationStack.Push(indentationStack.Peek() + indent);
      return result;
    }

    internal void InternalReduceIndentation() {
      indentationStack.Pop();
    }

    #endregion

    private void OnBeforePutNode(NodeItem item) {
      if (BeforePutNode != null) {
        BeforePutNode(this, new NodeEventArgs<TNode>(item.Node, this));
      }
      nodeStack.Push(item);
    }

    private void OnAfterPutNode(NodeItem item) {
      nodeStack.Pop();
      if (AfterPutNode != null) {
        AfterPutNode(this, new NodeEventArgs<TNode>(item.Node, this));
      }
    }

    private struct NodeItem {
      private readonly TNode node;
      private readonly object template;

      public NodeItem(TNode node, object template) {
        this.node = node;
        this.template = template;
      }

      public TNode Node {
        get { return node; }
      }
    }
  }
}