// Atac Template Engine
// Copyright (C) 2011 Markus Jahn <markus.jahn@gmx.at>
//
// This file is part of Atac Template Engine.
//
// Atac Template Engine is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Atac Template Engine is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Atac Template Engine.  If not, see <http://www.gnu.org/licenses/>.

using System.Collections;

namespace Atac.TemplateEngine {
  internal class ExpressionWriter<TNode> where TNode : class {
    private readonly CodeGenerator<TNode> codeGen;
    private Line currentLine;
    private bool isFirstLine = true;
    private bool isLinePending;
    private bool isLineDeletable;

    internal ExpressionWriter(CodeGenerator<TNode> codeGen) {
      this.codeGen = codeGen;
    }

    internal bool PutLine(Line line) {
      if (line == null) {
        return false;
      }
      currentLine = line;
      isLinePending = true;
      isLineDeletable = false;
      foreach (LineElement element in line.Elements) {
        element.Put(this);
      }
      if (isLinePending) {
        if (isLineDeletable) {
          return false;
        }
        PutPendingElements();
      }
      if (currentLine.Indentation != null) {
        codeGen.ReduceIndentation();
      }
      isFirstLine = false;
      return true;
    }

    internal void PutText(Text text) {
      PutPendingElements();
      codeGen.PutText(text.Value);
    }

    internal void PutPlaceholder(Placeholder placeholder) {
      isLineDeletable = true;
      object referencedObject = placeholder.Path != null
          ? codeGen.TreeHelper.GetPathValue(codeGen.CurrentNode, placeholder.Path) : null;
      if (referencedObject != null) {
        PutPendingElements();
        PutExpression(placeholder.Prefix);
        IEnumerator enumerator = YieldItem(referencedObject).GetEnumerator();
        if (enumerator.MoveNext()) {
          PutObject(enumerator.Current, placeholder.Link);
          while (enumerator.MoveNext()) {
            PutExpression(placeholder.Separator);
            PutObject(enumerator.Current, placeholder.Link);
          }
        }
        PutExpression(placeholder.Suffix);
      } else if (placeholder.Alternative != null && placeholder.Alternative.Lines.Count > 0) {
        PutPendingElements();
        PutExpression(placeholder.Prefix);
        PutExpression(placeholder.Alternative);
        PutExpression(placeholder.Suffix);
      }
    }

    private void PutExpression(Expression expression) {
      if (expression != null) {
        expression.Put(codeGen);
      }
    }

    private void PutObject(object value, string link) {
      if (value is TNode) {
        codeGen.InternalPutNode((TNode) value, link);
      } else {
        codeGen.InternalPutText(value);
      }
    }

    private void PutPendingElements() {
      if (!isLinePending) {
        return;
      }
      if (!isFirstLine) {
        codeGen.InternalPutNewLine(false);
      }
      if (currentLine.Indentation != null) {
        codeGen.InternalPutIndentation(currentLine.Indentation);
      }
      isLinePending = false;
    }


    private IEnumerable YieldItem(object member) {
      if (member is IEnumerable) {
        foreach (object value in (IEnumerable) member) {
          yield return value;
        }
      } else {
        yield return member;
      }
    }
  }
}