﻿// Atac Template Engine
// Copyright (C) 2011 Markus Jahn <markus.jahn@gmx.at>
//
// This file is part of Atac Template Engine.
//
// Atac Template Engine is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Atac Template Engine is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Atac Template Engine.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Text;

namespace Atac.TemplateEngine {
  public class Template {
    private readonly string name;
    private IList<PathDecl> pathDecls;
    private Expression expression;
    private int precedingLineFeeds;
    private int succeedingLineFeeds;

    public Template(string name) : this(name, new Expression()) {}

    public Template(string name, Expression expression)
        : this(name, expression, new PathDecl[0]) {}

    public Template(string name, IList<PathDecl> pathDecls)
      : this(name, new Expression(), pathDecls) { }

    public Template(string name, Expression expression, IList<PathDecl> pathDecls) {
      if (string.IsNullOrEmpty(name)) {
        throw new ArgumentException("Template name must not be null or empty", "name");
      }
      if (expression == null) {
        throw new ArgumentNullException("expression");
      }
      if (pathDecls == null) {
        throw new ArgumentNullException("pathDecls");
      }
      this.name = name;
      this.expression = expression;
      this.pathDecls = pathDecls;
    }

    public string Name {
      get { return name; }
    }
    public IList<PathDecl> PathDecls {
      get { return pathDecls; }
      set { pathDecls = value; }
    }
    public int PrecedingLineFeeds {
      get { return precedingLineFeeds; }
      set { precedingLineFeeds = value; }
    }
    public int SucceedingLineFeeds {
      get { return succeedingLineFeeds; }
      set { succeedingLineFeeds = value; }
    }
    public Expression Expression {
      get { return expression; }
      set { expression = value; }
    }

    public void Put<TNode>(TNode node, CodeGenerator<TNode> cg) where TNode : class {
      if (expression == null) {
        return;
      }
      int oldPendingLineFeeds = cg.PendingLineFeeds;
      int newPendingLineFeeds = PrecedingLineFeeds - cg.DumpedLineFeeds;
      if (newPendingLineFeeds > oldPendingLineFeeds) {
        cg.PendingLineFeeds = newPendingLineFeeds;
      }
      if (expression.Put(cg)) {
        int succeedingLineFeeds = SucceedingLineFeeds - cg.DumpedLineFeeds;
        if (succeedingLineFeeds > cg.PendingLineFeeds) {
          cg.PendingLineFeeds = succeedingLineFeeds;
        }
      } else {
        cg.PendingLineFeeds = oldPendingLineFeeds;
      }
    }

    public override string ToString() {
      StringBuilder sb = new StringBuilder();
      sb.Append(Name);
      if (pathDecls != null && pathDecls.Count > 0) {
        sb.Append("(");
        sb.Append(PathDecls[0]);
        for (int i = 1; i < PathDecls.Count; i++) {
          sb.Append(", ");
          sb.Append(PathDecls[i]);
        }
        sb.Append(")");
      }
      sb.AppendLine(" =");
      if (expression != null && expression.Lines.Count > 0) {
        sb.AppendLine(expression.ToString());
      }
      sb.AppendLine(".");
      return sb.ToString();
    }
  }
}