﻿// Atac Template Engine
// Copyright (C) 2011 Markus Jahn <markus.jahn@gmx.at>
//
// This file is part of Atac Template Engine.
//
// Atac Template Engine is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Atac Template Engine is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Atac Template Engine.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.CodeDom.Compiler;
using System.Collections.Generic;
using System.IO;
using System.Reflection;
using System.Text;

namespace Atac.TemplateEngine {
  public delegate void TemplateHandler<TNode>(TNode node, CodeGenerator<TNode> generator)
      where TNode : class;

  public class TemplateEngine<TNode> where TNode : class {
    private TreeHelper<TNode> treeHelper;

    private FileInfo templateFile, templateHandlerFile;
    private Stream templateStream, templateHandlerStream;
    private Type[] templateHandlerTypes;
    private object[] templateHandlerObjects;

    public Dictionary<string, Template> Templates { get; private set; }
    public Dictionary<string, TemplateHandler<TNode>> TemplateHandlers { get; private set; }
    public ILogger Logger { get; set; }

    public TreeHelper<TNode> TreeHelper {
      get { return treeHelper; }
      set {
        if (value == null) {
          throw new ArgumentNullException();
        }
        treeHelper = value;
      }
    }

    public TemplateEngine(string templateFile) {
      if (string.IsNullOrEmpty(templateFile)) {
        throw new ArgumentException("Template filename must not be null, empty, or consist only of whitespace", "templateFile");
      }
      Logger = new ConsoleLogger();
      TreeHelper = new MemberTreeHelper<TNode>();
      this.templateFile = new FileInfo(templateFile);
    }

    public TemplateEngine(string templateFile, string templateHandlerFile)
        : this(templateFile) {
      if (string.IsNullOrEmpty(templateHandlerFile)) {
        throw new ArgumentException("TemplateHandler filename must not be null, empty, or consist only of whitespace", "templateHandlerFile");
      }
      this.templateHandlerFile = new FileInfo(templateHandlerFile);
    }

    public TemplateEngine(string templateFile, Stream templateHandlerStream)
        : this(templateFile) {
      if (templateHandlerStream == null) {
        throw new ArgumentNullException("templateHandlerStream");
      }
      this.templateHandlerStream = templateHandlerStream;
    }

    public TemplateEngine(string templateFile, params Type[] templateHandlerTypes)
        : this(templateFile) {
      if (templateHandlerTypes == null) {
        throw new ArgumentNullException("templateHandlerTypes");
      }
      this.templateHandlerTypes = templateHandlerTypes;
    }

    public TemplateEngine(string templateFile, params object[] templateHandlerObjects)
        : this(templateFile) {
      if (templateHandlerObjects == null) {
        throw new ArgumentNullException("templateHandlerObjects");
      }
      this.templateHandlerObjects = templateHandlerObjects;
    }

    public TemplateEngine(Stream templateStream) {
      if (templateStream == null) {
        throw new ArgumentNullException("templateStream");
      }
      Logger = new ConsoleLogger();
      TreeHelper = new MemberTreeHelper<TNode>();
      this.templateStream = templateStream;
    }

    public TemplateEngine(Stream templateStream, string templateHandlerFile)
        : this(templateStream) {
      if (string.IsNullOrEmpty(templateHandlerFile)) {
        throw new ArgumentException("TemplateHandler file must not be null, empty, or whitespace", "templateHandlerFile");
      }
      this.templateHandlerFile = new FileInfo(templateHandlerFile);
    }

    public TemplateEngine(Stream templateStream, Stream templateHandlerStream)
        : this(templateStream) {
      if (templateHandlerStream == null) {
        throw new ArgumentNullException("templateHandlerStream");
      }
      this.templateHandlerStream = templateHandlerStream;
    }

    public TemplateEngine(Stream templateStream, params Type[] templateHandlerTypes)
        : this(templateStream) {
      if (templateHandlerTypes == null) {
        throw new ArgumentNullException("templateHandlerTypes");
      }
      this.templateHandlerTypes = templateHandlerTypes;
    }

    public TemplateEngine(Stream templateStream, params object[] templateHandlerObjects)
        : this(templateStream) {
      if (templateHandlerObjects == null) {
        throw new ArgumentNullException("templateHandlerObjects");
      }
      this.templateHandlerObjects = templateHandlerObjects;
    }

    public void Start(TextWriter writer, TNode root) {
      Start(writer, root, TreeHelper);
    }

    public void Start(TextWriter writer, TNode root, TreeHelper<TNode> treeHelper) {
      if (writer == null) {
        throw new ArgumentNullException("writer");
      }
      if (root == null) {
        throw new ArgumentNullException("root");
      }
      if (treeHelper == null) {
        throw new ArgumentNullException("treeHelper");
      }
      Init();
      CodeGenerator<TNode> cg = new CodeGenerator<TNode>(
          writer, treeHelper, Templates, TemplateHandlers, Logger);
      cg.PutNode(root);
      writer.Flush();
    }

    public void UpdateTemplates() {
      Templates = templateFile != null
          ? GetTemplatesFromFile(templateFile)
          : GetTemplatesFromStream(templateStream);
    }

    public void UpdateTemplates(string templateFile) {
      templateStream = null;
      if (string.IsNullOrEmpty(templateFile)) {
        throw new ArgumentException("Template filename must not be null, empty, or consist only of whitespace", "templateFile");
      }
      ResetTemplateSource();
      this.templateFile = new FileInfo(templateFile);
      Templates = GetTemplatesFromFile(this.templateFile);
    }

    public void UpdateTemplates(Stream templateStream) {
      if (templateStream == null) {
        throw new ArgumentNullException("templateStream");
      }
      ResetTemplateSource();
      this.templateStream = templateStream;
      Templates = GetTemplatesFromStream(templateStream);
    }

    public void UpdateTemplateHandlers() {
      if (templateHandlerFile != null) {
        TemplateHandlers = GetTemplateHandlersFromFile(templateHandlerFile);
      } else if (templateHandlerStream != null) {
        TemplateHandlers = GetTemplateHandlersFromStream(templateHandlerStream);
      } else if (templateHandlerTypes != null) {
        TemplateHandlers = GetTemplateHandlersFromTypes(templateHandlerTypes);
      } else if (templateHandlerObjects != null) {
        TemplateHandlers = GetTemplateHandlersFromObjects(templateHandlerObjects);
      } else {
        TemplateHandlers = new Dictionary<string, TemplateHandler<TNode>>();
      }
    }

    public void UpdateTemplateHandlers(string templateHandlerFile) {
      if (string.IsNullOrEmpty(templateHandlerFile)) {
        throw new ArgumentException("TemplateHandler filename must not be null, empty, or consist only of whitespace", "templateHandlerFile");
      }
      ResetTemplateHandlerSource();
      this.templateHandlerFile = new FileInfo(templateHandlerFile);
      TemplateHandlers = GetTemplateHandlersFromFile(this.templateHandlerFile);
    }

    public void UpdateTemplateHandlers(Stream templateHandlerStream) {
      if (templateHandlerStream == null) {
        throw new ArgumentNullException("templateHandlerStream");
      }
      ResetTemplateHandlerSource();
      this.templateHandlerStream = templateHandlerStream;
      TemplateHandlers = GetTemplateHandlersFromStream(templateHandlerStream);
    }


    public void UpdateTemplateHandlers(params Type[] templateHandlerTypes) {
      if (templateHandlerTypes == null) {
        throw new ArgumentNullException("templateHandlerTypes");
      }
      ResetTemplateHandlerSource();
      this.templateHandlerTypes = templateHandlerTypes;
      TemplateHandlers = GetTemplateHandlersFromTypes(templateHandlerTypes);
    }

    public void UpdateTemplateHandlers(params object[] templateHandlerObjects) {
      if (templateHandlerObjects == null) {
        throw new ArgumentNullException("templateHandlerObjects");
      }
      ResetTemplateHandlerSource();
      this.templateHandlerObjects = templateHandlerObjects;
      TemplateHandlers = GetTemplateHandlersFromObjects(templateHandlerObjects);
    }

    private void Init() {
      if (Templates == null) {
        UpdateTemplates();
      } else if (templateFile != null) {
        DateTime lastWriteTime = templateFile.LastWriteTime;
        templateFile.Refresh();
        if (lastWriteTime != templateFile.LastWriteTime) {
          UpdateTemplates();
        }
      }
      if (TemplateHandlers == null) {
        UpdateTemplateHandlers();
      } else if (templateHandlerFile != null) {
        DateTime lastWriteTime = templateHandlerFile.LastWriteTime;
        templateHandlerFile.Refresh();
        if (lastWriteTime != templateHandlerFile.LastWriteTime) {
          UpdateTemplateHandlers();
        }
      }
    }

    private void ResetTemplateSource() {
      templateFile = null;
      templateStream = null;
    }

    private void ResetTemplateHandlerSource() {
      templateHandlerFile = null;
      templateHandlerStream = null;
      templateHandlerTypes = null;
      templateHandlerObjects = null;
    }

    private Dictionary<string, Template> GetTemplatesFromFile(FileInfo file) {
      Dictionary<string, Template> templates;
      file.Refresh();
      if (!file.Exists) {
        if (Logger != null) {
          Logger.LogError(string.Format("Template file '{0}' does not exist", file.FullName));
        }
        templates = new Dictionary<string, Template>();
      } else {
        using (FileStream fs = file.Open(FileMode.Open)) {
          templates = GetTemplatesFromStream(fs);
        }
      }
      return templates;
    }

    private Dictionary<string, Template> GetTemplatesFromStream(Stream stream) {
      TemplateParser parser = new TemplateParser(new Scanner(stream), Logger);
      parser.Parse();
      return parser.Templates;
    }

    private Dictionary<string, TemplateHandler<TNode>> GetTemplateHandlersFromFile(FileInfo handlerFile) {
      Dictionary<string, TemplateHandler<TNode>> templateHandlers;
      using (FileStream fs = handlerFile.Open(FileMode.Open)) {
        templateHandlers = GetTemplateHandlersFromStream(fs);
        handlerFile.Refresh();
      }
      return templateHandlers;
    }

    private Dictionary<string, TemplateHandler<TNode>> GetTemplateHandlersFromStream(Stream handlerStream) {
      CodeDomProvider cdp = CodeDomProvider.CreateProvider(
          CodeDomProvider.GetLanguageFromExtension("cs"));
      CompilerParameters cp = new CompilerParameters();
      cp.IncludeDebugInformation = true;
      cp.GenerateExecutable = false;
      cp.GenerateInMemory = true;
      cp.ReferencedAssemblies.Add(Assembly.GetExecutingAssembly().Location);

      StringBuilder sb = new StringBuilder();

      StreamReader reader = new StreamReader(handlerStream);

      string line = reader.ReadLine();
      bool referencesAdded = false;
      while (line != null && !referencesAdded) {
        sb.AppendLine(line);
        if (line.StartsWith("//r:")) {
          cp.ReferencedAssemblies.Add(line.Substring(4).Trim());
        } else if (line.StartsWith("//reference:")) {
         cp.ReferencedAssemblies.Add(line.Substring(12).Trim());
        } else {
          referencesAdded = true;
        }
        line = reader.ReadLine();
      }
      if (line != null) {
        sb.AppendLine(line);
        sb.Append(reader.ReadToEnd());
      }
      CompilerResults results = cdp.CompileAssemblyFromSource(cp, sb.ToString());
      if (results.Errors.HasErrors) {
        if (Logger != null) {
          foreach (CompilerError error in results.Errors) {
            if (error.IsWarning) {
              Logger.LogWarning(string.Format("Warning on creating template handler object: {0}", error.ErrorText),
                  error.Line, error.Column);
            } else {
              Logger.LogError(string.Format("Error on creating template handler object: {0}", error.ErrorText),
                  error.Line, error.Column);
            }
          }
        }
        return new Dictionary<string, TemplateHandler<TNode>>();
      }
      return GetTemplateHandlersFromTypes(results.CompiledAssembly.GetTypes());
    }

    private Dictionary<string, TemplateHandler<TNode>> GetTemplateHandlersFromTypes(IEnumerable<Type> handlerTypes) {
      List<object> handlerObjects = new List<object>();
      foreach (Type type in handlerTypes) {
        try {
          handlerObjects.Add(Activator.CreateInstance(type));
        } catch (Exception e) {
          Logger.LogError(e.ToString());
        }
      }
      return GetTemplateHandlersFromObjects(handlerObjects);
    }

    private Dictionary<string, TemplateHandler<TNode>> GetTemplateHandlersFromObjects(IEnumerable<object> handlerObjects) {
      Dictionary<string, TemplateHandler<TNode>> templateHandlers = new Dictionary<string, TemplateHandler<TNode>>();
      foreach (object templateHandlerObject in handlerObjects) {
        Type type = templateHandlerObject.GetType();
        MethodInfo[] handlerCandidates = type.GetMethods(
            BindingFlags.Static | BindingFlags.Instance | BindingFlags.NonPublic | BindingFlags.Public);
        foreach (MethodInfo method in handlerCandidates) {
          if (method.ReturnType != typeof(void)) {
            continue;
          }
          ParameterInfo[] parameters = method.GetParameters();
          if (parameters.Length != 2
              || !typeof(TNode).IsAssignableFrom(parameters[0].ParameterType)
                  || !typeof(CodeGenerator<TNode>).IsAssignableFrom(parameters[1].ParameterType)) {
            continue;
          }
          if (templateHandlers.ContainsKey(method.Name)) {
            if (Logger != null) {
              Logger.LogWarning(string.Format("Template handler '{0}' is already in template handler collection", method.Name));
            }
            continue;
          }
          TemplateHandler<TNode> templateHandler = (TemplateHandler<TNode>) Delegate.CreateDelegate(
              typeof(TemplateHandler<TNode>), templateHandlerObject, method, false);
          if (templateHandler != null) {
            templateHandlers.Add(method.Name, templateHandler);
          }
        }
      }
      return templateHandlers;
    }
  }
}