﻿using System;
using System.Collections.Generic;

namespace Taste {
  public enum Type {
    UNDEF,
    INT,
    BOOL
  }
  public enum Operator {
    EQU,
    LSS,
    GTR,
    ADD,
    SUB,
    MUL,
    DIV
  }

  public abstract class Node {
    // node of the AST
    public abstract void Dump();
  }

  //----------- Declarations ----------------------------

  public class Obj : Node {
    // any declared object that has a name
    public String name; // name of this object
    public Type type; // type of this object (UNDEF for procedures)

    public Obj(String s, Type t) {
      name = s;
      type = t;
    }

    public override void Dump() {}
  }

  public class Var : Obj {
    // variable
    public int adr; // address in memory

    public Var(String name, Type type) : base(name, type) { }
  }

  public class Proc : Obj {
    // procedure (also used for the whole program)
    public List<Obj> locals; // objects declared in this scope
    public Block block; // block of this scope (null for the main program)
    public int nextAdr; // next free address in this scope
    public Proc program; // link to the Proc node of the main program or null
    public Parser parser; // for error messages

    public Proc(String name, Proc program, Parser parser) : base(name, Type.UNDEF) {
      locals = new List<Obj>();
      this.program = program;
      this.parser = parser;
    }

    public void Add(Obj obj) {
      foreach (Obj o in locals) {
        if (o.name.Equals(obj.name)) {
          parser.SemErr(obj.name + " declared twice");
        }
      }
      locals.Add(obj);
      if (obj is Var) {
        ((Var) obj).adr = nextAdr++;
      }
    }

    public Obj Find(String name) {
      foreach (Obj o in locals) {
        if (o.name.Equals(name)) {
          return o;
        }
      }
      if (program != null) {
        foreach (Obj o in program.locals) {
          if (o.name.Equals(name)) {
            return o;
          }
        }
      }
      parser.SemErr(name + " undeclared");
      return new Obj("_undef", Type.INT); // dummy
    }

    public override void Dump() {
      Console.WriteLine("Proc " + name);
      block.Dump();
      Console.WriteLine();
    }
  }

//----------- Expressions ------------------------------

  public abstract class Expr : Node {}

  public class BinExpr : Expr {
    public Operator op;
    public Expr left, right;

    public BinExpr(Expr e1, Operator o, Expr e2) {
      op = o;
      left = e1;
      right = e2;
    }

    public override void Dump() {
      left.Dump();
      Console.Write(" " + op + " ");
      right.Dump();
    }
  }

  public class UnaryExpr : Expr {
    public Operator op;
    public Expr e;

    public UnaryExpr(Operator x, Expr y) {
      op = x;
      e = y;
    }

    public override void Dump() {
      Console.Write(op + " ");
      e.Dump();
    }
  }

  public class Ident : Expr {
    public Obj obj;

    public Ident(Obj o) {
      obj = o;
    }

    public override void Dump() {
      Console.Write(obj.name);
    }
  }

  public class IntCon : Expr {
    public int val;

    public IntCon(int x) {
      val = x;
    }

    public override void Dump() {
      Console.Write(val);
    }
  }

  public class BoolCon : Expr {
    public bool val;

    public BoolCon(bool x) {
      val = x;
    }

    public override void Dump() {
      Console.Write(val);
    }
  }

//------------- Statements -----------------------------

  public class Stat : Node {
    public static int indent = 0;

    public override void Dump() {
      for (int i = 0; i < indent; i++) {
        Console.Write("  ");
      }
    }
  }

  public class Assignment : Stat {
    public Obj left;
    public Expr right;

    public Assignment(Obj o, Expr e) {
      left = o;
      right = e;
    }

    public override void Dump() {
      base.Dump();

      Console.Write(left.name + " = ");
      right.Dump();

      Console.WriteLine();
    }
  }

  public class Call : Stat {
    public Obj proc;

    public Call(Obj o) {
      proc = o;
    }

    public override void Dump() {
      base.Dump();

      Console.WriteLine("call " + proc.name);
    }
  }

  public class If : Stat {
    public Expr cond;
    public Stat stat;

    public If(Expr e, Stat s) {
      cond = e;
      stat = s;
    }

    public override void Dump() {
      base.Dump();
      Console.Write("if ");
      cond.Dump();
      Console.WriteLine();
      Stat.indent++;
      stat.Dump();
      Stat.indent--;
    }
  }

  public class IfElse : Stat {
    public Stat ifPart;
    public Stat elsePart;

    public IfElse(Stat i, Stat e) {
      ifPart = i;
      elsePart = e;
    }

    public override void Dump() {
      ifPart.Dump();
      base.Dump();

      Console.WriteLine("else ");
      Stat.indent++;
      elsePart.Dump();
      Stat.indent--;
    }
  }

  public class While : Stat {
    public Expr cond;
    public Stat stat;

    public While(Expr e, Stat s) {
      cond = e;
      stat = s;
    }

    public override void Dump() {
      base.Dump();

      Console.Write("while ");
      cond.Dump();

      Console.WriteLine();
      Stat.indent++;
      stat.Dump();
      Stat.indent--;
    }
  }

  public class Read : Stat {
    public Obj obj;

    public Read(Obj o) {
      obj = o;
    }

    public override void Dump() {
      base.Dump();

      Console.WriteLine("read " + obj.name);
    }
  }

  public class Write : Stat {
    public Expr e;

    public Write(Expr x) {
      e = x;
    }

    public override void Dump() {
      base.Dump();

      Console.Write("write ");
      e.Dump();

      Console.WriteLine();
    }
  }

  public class Block : Stat {
    public List<Stat> stats = new List<Stat>();

    public void Add(Stat s) {
      stats.Add(s);
    }

    public override void Dump() {
      base.Dump();

      Console.WriteLine("Block(");
      Stat.indent++;
      foreach (Stat s in stats) {
        s.Dump();
      }
      Stat.indent--;
      base.Dump();

      Console.WriteLine(")");
    }
  }
}